--                                                                    --
--  MAX Home Automation             Copyright (c)  Dmitry A. Kazakov  --
--     MAX_MQTT_Server                             Luebeck            --
--  Implementation                                 Spring, 2016       --
--                                                                    --
--                                Last revision :  19:17 30 Jan 2020  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--

with Ada.IO_Exceptions;      use Ada.IO_Exceptions;
with GNAT.Sockets.MQTT;      use GNAT.Sockets.MQTT;
with Strings_Edit.Floats;    use Strings_Edit.Floats;
with Strings_Edit.Integers;  use Strings_Edit.Integers;

package body MAX_MQTT_Server is

   Radiator_Messages  : MQTT_Messages_Array (1..7);
   Shutter_Messages   : MQTT_Messages_Array (1..4);
   Wall_Messages      : MQTT_Messages_Array (1..6);
   Statistic_Messages : MQTT_Messages_Array (1..4);

   function Image (Mode : Operating_Mode) return String is
   begin
      case Mode is
         when Automatic =>
            return "automatic";
         when Manual =>
            return "manual";
         when Boost =>
            return "boost";
         when Vacation =>
            return "vacation";
      end case;
   end Image;

   procedure Publish
             (  Client    : in out ELV_MAX_Cube_Client'Class;
                Server    : in out MQTT_Server'Class;
                Cube      : RF_Address;
                Connected : Boolean
             )  is
      Topic : constant String := Image (Cube) & "/connection";
   begin
      if Connected then
         Server.Publish (Topic, "on", Policy => Updated);
      else
         Server.Publish (Topic, "off", Policy => Updated);
      end if;
   end Publish;

   procedure Publish
             (  Client : in out ELV_MAX_Cube_Client'Class;
                Server : in out MQTT_Server'Class;
                Data   : Device_Data
             )  is
      procedure Do_Battery
                (  Message : in out MQTT_Message;
                   Prefix  : String
                )  is
      begin
         Set_Topic (Message, Prefix & "battery");
         if Data.Battery_Low then
            Set_Message (Message, "low");
         else
            Set_Message (Message, "high");
         end if;
      end Do_Battery;

      procedure Do_Error
                (  Message : in out MQTT_Message;
                   Prefix  : String
                )  is
      begin
         Set_Topic (Message, Prefix & "error");
         if Data.Error then
            Set_Message (Message, "error");
         else
            Set_Message (Message, "ok");
         end if;
      end Do_Error;

      procedure Do_Link
                (  Message : in out MQTT_Message;
                   Prefix  : String
                )  is
      begin
         Set_Topic (Message, Prefix & "link");
         if Data.Link_Error then
            Set_Message (Message, "error");
         else
            Set_Message (Message, "ok");
         end if;
      end Do_Link;

      procedure Do_Mode
                (  Message : in out MQTT_Message;
                   Prefix  : String
                )  is
      begin
         Set_Topic (Message, Prefix & "mode");
         Set_Message (Message, Image (Data.Mode));
      end Do_Mode;

      procedure Do_Set_Temperature
                (  Message : in out MQTT_Message;
                   Prefix  : String
                )  is
      begin
         Set_Topic (Message, Prefix & "set temperature");
         Set_Message (Message, Image (Data.Set_Temperature));
      end Do_Set_Temperature;

      procedure Do_Temperature
                (  Message : in out MQTT_Message;
                   Prefix  : String
                )  is
      begin
         Set_Topic (Message, Prefix & "temperature");
         Set_Message (Message, Image (Data.Temperature));
      end Do_Temperature;

      procedure Do_Valve
                (  Message : in out MQTT_Message;
                   Prefix  : String
                )  is
      begin
         Set_Topic (Message, Prefix & "valve");
         Set_Message
         (  Message,
            Image (Integer (Float (Data.Valve_Position) * 100.0))
         );
      end Do_Valve;

   begin
      case Data.Kind_Of is
         when Shutter_Contact =>
            declare
               Root : constant String :=
                               Image (Client.Get_RF_Address) &
                               "/shutter contact/" &
                               Image (Data.Address) &
                               '/';
            begin
               Do_Battery (Shutter_Messages (1), Root);
               Set_Topic (Shutter_Messages (2), Root & "status");
               if Data.Open then
                  Set_Message (Shutter_Messages (2), "open");
               else
                  Set_Message (Shutter_Messages (2), "close");
               end if;
               Do_Link  (Shutter_Messages (3), Root);
               Do_Error (Shutter_Messages (4), Root);
               Server.Publish (Shutter_Messages, Policy => Updated);
            end;
         when Radiator_Thermostat | Radiator_Thermostat_Plus =>
            declare
               Root : constant String :=
                               Image (Client.Get_RF_Address) &
                               "/radiator thermostat/" &
                               Image (Data.Address) &
                               '/';
            begin
               Do_Battery (Radiator_Messages (1), Root);
               Do_Mode (Radiator_Messages (2), Root);
               Do_Set_Temperature (Radiator_Messages (3), Root);
               Do_Valve (Radiator_Messages (4), Root);
               Do_Link  (Radiator_Messages (5), Root);
               Do_Error (Radiator_Messages (6), Root);
               if Data.Temperature = Centigrade'First then
                  Server.Publish
                  (  Messages => Radiator_Messages (1..7),
                     Policy   => Updated
                  );
               else
                  Do_Temperature (Radiator_Messages (7), Root);
                  Server.Publish (Radiator_Messages, Policy => Updated);
               end if;
            end;
         when Wall_Thermostat =>
            declare
               Root : constant String :=
                               Image (Client.Get_RF_Address) &
                               "/wall thermostat/" &
                               Image (Data.Address) &
                               '/';
            begin
               Do_Battery (Wall_Messages (1), Root);
               Do_Mode (Wall_Messages (2), Root);
               Do_Set_Temperature (Wall_Messages (3), Root);
               Do_Temperature (Wall_Messages (4), Root);
               Do_Link  (Wall_Messages (5), Root);
               Do_Error (Wall_Messages (6), Root);
               Server.Publish (Wall_Messages, Policy => Updated);
            end;
         when others =>
            null;
      end case;
   exception
      when Status_Error =>
         null;
   end Publish;

   procedure Publish
             (  Client  : in out ELV_MAX_Cube_Client'Class;
                Server  : in out MQTT_Server'Class;
                Cube    : RF_Address;
                Average : Integer;
                Min     : Integer;
                Max     : Integer
             )  is
      Root : constant String := Image (Client.Get_RF_Address) &
                                "/thermostat valves/";
   begin
      Set_Topic (Statistic_Messages (1), Root & "max");
      Set_Message (Statistic_Messages (1), Image (Max));
      Set_Topic (Statistic_Messages (2), Root & "min");
      Set_Message (Statistic_Messages (2), Image (Min));
      Set_Topic (Statistic_Messages (3), Root & "average");
      Set_Message (Statistic_Messages (3), Image (Average));
      Server.Publish (Statistic_Messages, Policy => Updated);
   end Publish;

   procedure Publish
             (  Client : in out ELV_MAX_Cube_Client'Class;
                Server : in out MQTT_Server'Class;
                Cube   : RF_Address;
                Duty   : Ratio
             )  is
   begin
      Server.Publish
      (  Topic   => Image (Cube) & "/duty",
         Message => Image (Float (Duty), AbsSmall => -2),
         Policy  => Updated
      );
   end Publish;

end MAX_MQTT_Server;
