--                                                                    --
--  package Py                      Copyright (c)  Dmitry A. Kazakov  --
--  Implementation                                 Luebeck            --
--                                                 Winter, 2018       --
--                                                                    --
--                                Last revision :  19:02 31 Dec 2018  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--

with Ada.IO_Exceptions;           use Ada.IO_Exceptions;
with Strings_Edit;                use Strings_Edit;
with Strings_Edit.Quoted;         use Strings_Edit.Quoted;
with Strings_Edit.UTF8;           use Strings_Edit.UTF8;
with Strings_Edit.UTF8.Handling;  use Strings_Edit.UTF8.Handling;

with Py.Load_Python_Library;

package body Py is
--
-- Python library bug work-around.  The exported  exception  objects are
-- broken. The true object will be detected by other means.
--
   TypeError   : Object := Null_Object;
   SystemError : Object := Null_Object;

   procedure Load (Name : String := "") is
   begin
      Py.Load_Python_Library.Load (Name);
   end Load;

   function As_String (Value : Object) return String is
      Length : ssize_t;
   begin
      Length := Links.Unicode_GetLength (Value);
      if Length = 0 then
         return "";
      else
         declare
            Result  : String (1..Integer (Length * 4));
            Pointer : Integer := 1;
         begin
            for Index in 0..Length - 1 loop
               Put
               (  Result,
                  Pointer,
                  Code_Point
                  (  Wide_Wide_Character'Pos
                     (  Links.Unicode_ReadChar (Value, Index)
               )  )  );
            end loop;
            return Result (1..Pointer - 1);
         end;
      end if;
   end As_String;

   function As_String (Object : Handle) return String is
   begin
      if Object.Ptr = Null_Object then
         return "";
      else
         return As_String (Object.Ptr);
      end if;
   end As_String;

   function Bytes_AsString (Object : Handle) return String is
      use Char_Ptrs;
      Ptr  : aliased Pointer;
      Size : aliased ssize_t := 0;
   begin
      if (  0 > Links.Bytes_AsStringAndSize
                (  Object.Ptr,
                   Ptr'Access,
                   Size'Access
                )
         or else
            Ptr = null
         or else
            Size = 0
         )
      then
         return "";
      else
         declare
            Result : String (1..Integer (Size));
         begin
            for Index in Result'Range loop
               Result (Index) := To_Ada (Ptr.all);
               Increment (Ptr);
            end loop;
            return Result;
         end;
      end if;
   end Bytes_AsString;

   function Callable_Check (Object : Handle) return int is
   begin
      return Links.Callable_Check (Object.Ptr);
   end Callable_Check;

   procedure Check_Error is
   begin
      if Null_Object /= Links.Err_Occurred.all then
         declare
            Error_Type  : Object;
            Error_Value : Object;
            Traceback   : Object;
            function Message return String is
               Encoded : Object;
               Ptr     : aliased Char_Ptrs.Pointer;
               Size    : aliased ssize_t;
            begin
               if Error_Value = Null_Object then
                  return "[No error message]";
               end if;
               Encoded := Links.Unicode_AsEncodedString (Error_Value);
               if Encoded = Null_Object then
                  return "[Malcoded error message]";
               end if;
               if ( -1
                  /= Links.Bytes_AsStringAndSize
                     (  Encoded,
                        Ptr'Access,
                        Size'Access
                  )  )  then
                  declare
                     Result : constant String :=
                                       To_Ada
                                       (  Char_Ptrs.Value
                                          (  Ptr,
                                             ptrdiff_t (Size)
                                          ),
                                          False
                                       );
                  begin
                     Links.DecRef (Encoded);
                     return Result;
                  end;
               else
                  Links.DecRef (Encoded);
                  return "[Masformed message]";
               end if;
            end Message;
         begin
            Links.Err_Fetch (Error_Type, Error_Value, Traceback);
            declare
               Text : constant String := Message;
            begin
               Links.Err_Restore (Error_Type, Error_Value, Traceback);
               Raise_Exception (Python_Error'Identity, Text);
            end;
         end;
      end if;
   end Check_Error;

   procedure Check_Links (Library_Name : String) is
      procedure Error (Entry_Name : String) is
      begin
         Raise_Exception
         (  Use_Error'Identity,
            (  "Failed to locate entry "
            &  Entry_Name
            &  " in the Python DLL "
            &  Quote (Library_Name)
         )  );
      end Error;
   begin
      if Links.AddPendingCall = null then
         Error ("Py_AddPendingCall");
      elsif Links.Bytes_AsStringAndSize = null then
         Error ("PyBytes_AsStringAndSize");
      elsif Links.Callable_Check = null then
         Error ("PyCallable_Check");
      elsif Links.DecRef = null then
         Error ("Py_DecRef");
      elsif Links.Dict_GetItemString = null then
         Error ("Py_Dict_GetItemString");
      elsif Links.Dict_Keys = null then
         Error ("Py_Dict_Keys");
      elsif Links.Dict_Items = null then
         Error ("Py_Dict_Items");
      elsif Links.Dict_New = null then
         Error ("Py_Dict_New");
      elsif Links.Dict_SetItemString = null then
         Error ("Py_Dict_SetItemString");
      elsif Links.Dict_Size = null then
         Error ("Py_Dict_Size");
      elsif Links.Err_BadArgument = null then
         Error ("PyErr_BadArgument");
      elsif Links.Err_BadInternalCall = null then
         Error ("PyErr_BadInternalCall");
      elsif Links.Err_Clear = null then
         Error ("PyErr_Clear");
      elsif Links.Err_Fetch = null then
         Error ("PyErr_Fetch");
      elsif Links.Err_NewException = null then
         Error ("PyErr_NewException");
      elsif Links.Err_NormalizeException = null then
         Error ("PyErr_NormalizeException");
      elsif Links.Err_Occurred = null then
         Error ("PyErr_Occurred");
      elsif Links.Err_SetString = null then
         Error ("PyErr_SetString");
      elsif Links.Eval_InitThreads = null then
         Error ("PyEval_InitThreads");
      elsif Links.Exc_KeyError = Null_Object then
         Error ("PyExc_KeyError");
      elsif Links.Exc_RuntimeError = Null_Object then
         Error ("PyExc_RuntimeError");
      elsif Links.Exc_SystemError = Null_Object then
         Error ("PyExc_SystemError");
      elsif Links.Exc_TypeError = Null_Object then
         Error ("PyExc_TypeError");
      elsif Links.Exc_ValueError = Null_Object then
         Error ("PyExc_ValueError");
      elsif Links.FinalizeEx = null then
         Error ("Py_FinalizeEx");
      elsif Links.GILState_Ensure = null then
         Error ("PyGILState_Ensure");
      elsif Links.GILState_Release = null then
         Error ("PyGILState_Release");
      elsif Links.Float_AsDouble = null then
         Error ("PyFloat_AsDouble");
      elsif Links.Float_FromDouble = null then
         Error ("PyFloat_FromDouble");
      elsif Links.Import_AppendInittab = null then
         Error ("PyImport_AppendInittab");
      elsif Links.Import_Import = null then
         Error ("PyImport_Import");
      elsif Links.Import_ImportModule = null then
         Error ("PyImport_ImportModule");
      elsif Links.IncRef = null then
         Error ("Py_IncRef");
      elsif Links.InitializeEx = null then
         Error ("Py_InitializeEx");
      elsif Links.List_Append = null then
         Error ("PyList_Append");
      elsif Links.List_Insert = null then
         Error ("PyList_Insert");
      elsif Links.List_GetItem = null then
         Error ("PyList_GetItem");
      elsif Links.List_New = null then
         Error ("PyList_New");
      elsif Links.List_Size = null then
         Error ("PyList_Size");
      elsif Links.Long_AsLong = null then
         Error ("PyLong_AsLong");
      elsif Links.Long_FromLong = null then
         Error ("PyLong_FromLong");
      elsif Links.Module_Create = null then
         Error ("PyModule_Create2");
      elsif Links.None = Null_Object then
         Error ("_PyNone_Struct");
      elsif Links.Object_CallObject = null then
         Error ("PyObject_CallObject");
      elsif Links.Object_GetAttrString = null then
         Error ("PyObject_GetAttrString");
      elsif Links.Object_Dir = null then
         Error ("PyObject_Dir");
      elsif Links.Object_Str = null then
         Error ("PyObject_Str");
      elsif Links.Sequence_DelItem = null then
         Error ("PySequence_DelItem");
      elsif Links.Sequence_Index = null then
         Error ("PySequence_Index");
      elsif Links.Sequence_Index = null then
         Error ("PySys_GetObject");
      elsif Links.Unicode_AsEncodedString = null then
         Error ("PyUnicode_AsEncodedString");
      elsif Links.Tuple_GetItem = null then
         Error ("PyTuple_GetItem");
      elsif Links.Tuple_New = null then
         Error ("PyTuple_New");
      elsif Links.Tuple_SetItem = null then
         Error ("PyTuple_SetItem");
      elsif Links.Tuple_Size = null then
         Error ("PyTuple_Size");
      elsif Links.Unicode_GetLength = null then
         Error ("Unicode_GetLength");
      elsif Links.Unicode_DecodeFSDefault = null then
         Error ("PyUnicode_DecodeFSDefault");
      elsif Links.Unicode_FromString = null then
         Error ("Unicode_FromString");
      elsif Links.Unicode_ReadChar = null then
         Error ("Unicode_ReadChar");
      end if;
   end Check_Links;

   procedure Check_Handle (Object : Handle) is
      pragma Inline (Check_Handle);
   begin
      if Object.Ptr = Null_Object then
         Raise_Exception (Constraint_Error'Identity, "Null object");
      end if;
   end Check_Handle;

   function Dict_GetItemString
            (  Dictionary : Handle;
               Key        : String
            )  return Handle is
      Result : Handle;
   begin
      Check_Handle (Dictionary);
      Result.Ptr :=
         Links.Dict_GetItemString (Dictionary.Ptr, To_C (Key));
      if Result.Ptr /= Null_Object then -- Borrowed reference
         Links.IncRef (Result.Ptr);
      end if;
      return Result;
   end Dict_GetItemString;

   function Dict_Keys (Dictionary : Handle) return Handle is
      Result : Handle;
   begin
      Check_Handle (Dictionary);
      Result.Ptr := Links.Dict_Keys (Dictionary.Ptr);
      if Result.Ptr = Null_Object then -- New reference
         Check_Error;
      end if;
      return Result;
   end Dict_Keys;

   function Dict_Items (Dictionary : Handle) return Handle is
      Result : Handle;
   begin
      Check_Handle (Dictionary);
      Result.Ptr := Links.Dict_Items (Dictionary.Ptr);
      if Result.Ptr = Null_Object then -- New reference
         Check_Error;
      end if;
      return Result;
   end Dict_Items;

   function Dict_New return Handle is
   begin
      return (Ada.Finalization.Controlled with Links.Dict_New.all);
   end Dict_New;

   procedure Dict_SetItemString
             (  Dictionary : Handle;
                Key        : String;
                Value      : Handle'Class
             ) is
   begin
      if 0 /= Links.Dict_SetItemString
              (  Dictionary.Ptr,
                 To_C (Key),
                 Value.Ptr
              )
      then
         Raise_Exception
         (  Data_Error'Identity,
            "Failed to set dictionary item"
         );
      end if;
   end Dict_SetItemString;

   function Dict_Size (Dictionary : Handle) return ssize_t is
      Result : ssize_t;
   begin
      Check_Handle (Dictionary);
      Result := Links.Dict_Size (Dictionary.Ptr);
      Check_Error;
      return Result;
   end Dict_Size;

   procedure Err_Clear is
   begin
      Links.Err_Clear.all;
   end Err_Clear;

   function Err_Occurred return Handle is
   begin
      return (Ada.Finalization.Controlled with Links.Err_Occurred.all);
   end Err_Occurred;

   procedure Eval_InitThreads is
   begin
      Links.Eval_InitThreads.all;
   end Eval_InitThreads;

   procedure Finalize (Lock : in out Global_Interpreter_Lock) is
   begin
      Links.GILState_Release (Lock.State);
   end Finalize;

   function FinalizeEx return int is
   begin
      TypeError   := Null_Object;
      SystemError := Null_Object;
      return Links.FinalizeEx.all;
   end FinalizeEx;

   procedure Initialize is
   begin
      Links.InitializeEx (0);
   end Initialize;

   function Import_AppendInittab
            (  Name : char_array;
               Init : InitTab
            )  return int is
   begin
      return Links.Import_AppendInittab (Name, Init);
   end Import_AppendInittab;

   function Import_Import (Name : Handle) return Handle is
      Result : Handle;
   begin
      Check_Handle (Name);
      Result.Ptr := Links.Import_Import (Name.Ptr);
      if Result.Ptr = Null_Object then
         Check_Error;
      end if;
      return Result;
   end Import_Import;

   procedure Import_Import
             (  Name        : Handle;
                Module      : out Handle;
                Destination : in out String;
                Pointer     : in out Integer;
                No_Error    : out Boolean;
                Decorator   : String := Character'Val (13) &
                                        Character'Val (10) &
                                        "   "
             )  is
   begin
      Initialize (Module);
      Check_Handle (Name);
      Module.Ptr := Links.Import_Import (Name.Ptr);
      if (  Module.Ptr = Null_Object
         or else
            Null_Object /= Links.Err_Occurred.all
         )  then
         Error_Traceback (Destination, Pointer, No_Error, Decorator);
      else
         No_Error := True;
      end if;
   end Import_Import;

   procedure Import_Import
             (  Name        : String;
                Module      : out Handle;
                Destination : in out String;
                Pointer     : in out Integer;
                No_Error    : out Boolean;
                Decorator   : String := Character'Val (13) &
                                        Character'Val (10) &
                                        "   "
             )  is
   begin
      Initialize (Module);
      Module.Ptr := Links.Import_ImportModule (To_C (Name));
      if (  Module.Ptr = Null_Object
         or else
            Null_Object /= Links.Err_Occurred.all
         )  then
         Error_Traceback (Destination, Pointer, No_Error, Decorator);
      else
         No_Error := True;
      end if;
   end Import_Import;

   function Import_ImportModule (Name : String) return Handle is
      Result : Handle;
   begin
      Result.Ptr := Links.Import_ImportModule (To_C (Name));
      if Result.Ptr = Null_Object then
         Check_Error;
      end if;
      return Result;
   end Import_ImportModule;

   procedure Initialize (Lock : in out Global_Interpreter_Lock) is
   begin
      Lock.State := Links.GILState_Ensure.all;
   end Initialize;

   procedure Invalidate (Object : in out Handle) is
   begin
      if Object.Ptr /= Null_Object then
         Links.DecRef (Object.Ptr);
         Object.Ptr := Null_Object;
      end if;
   end Invalidate;

   function Is_None (Object : Handle) return Boolean is
   begin
      return Object.Ptr = Null_Object or else Object.Ptr = Links.None;
   end Is_None;

   function Is_Null (Reference : Object) return Boolean is
   begin
      return Reference = Null_Object;
   end Is_Null;

   function Is_Valid (Object : Handle) return Boolean is
   begin
      return Object.Ptr /= Null_Object;
   end Is_Valid;

   procedure Adjust (Object : in out Handle) is
   begin
      if Object.Ptr /= Null_Object then
         Links.IncRef (Object.Ptr);
      end if;
   end Adjust;

   procedure Error_Traceback
             (  Destination : in out String;
                Pointer     : in out Integer;
                No_Error    : out Boolean;
                Decorator   : String := Character'Val (13) &
                                        Character'Val (10) &
                                        "   "
             )  is
      Error : Object;
   begin
      Error := Links.Err_Occurred.all;
      if Error = Null_Object then
         No_Error := True;
         return;
      end if;
      No_Error := False;
      declare
         Class   : Handle;
         Message : Handle;
         Trace   : Handle;
      begin
         Links.Err_Fetch (Class.Ptr, Message.Ptr, Trace.Ptr);
         Links.Err_NormalizeException
         (  Class.Ptr,
            Message.Ptr,
            Trace.Ptr
         );
         Put (Destination, Pointer, Object_Str (Message));
         Put (Destination, Pointer, " [");
         Put (Destination, Pointer, Object_Str (Class));
         Put (Destination, Pointer, "]");
         if not Is_Valid (Trace) then
            return;
         end if;
         declare
            Func   : Handle;
            Module : Handle;
         begin
            Module :=
               Import_Import (Unicode_DecodeFSDefault ("traceback"));
            Func := Object_GetAttrString (Module, "format_tb");
            if Callable_Check (Func) = 0 then
               declare
                  Args  : Handle;
                  Value : Handle;
               begin
                  Args := Tuple_New (1);
                  Tuple_SetItem (Args, 0, Trace);
                  Value := Object_CallObject (Func, Args);
                  if Is_Valid (Value) then
                     declare
                        Length : constant ssize_t := List_Size (Value);
                        Item   : Object;
                     begin
                        for Index in 0..Length - 1 loop
                           Item :=
                              Links.List_GetItem (Value.Ptr, Index);
                           if Item /= Null_Object then
                              Put (Destination, Pointer, Decorator);
                              Put
                              (  Destination,
                                 Pointer,
                                 As_String (Item)
                              );
                           end if;
                        end loop;
                     end;
                  end if;
               end;
            end if;
         end;
      end;
   end Error_Traceback;

   procedure Finalize (Object : in out Handle) is
   begin
      Invalidate (Object);
   end Finalize;

   function Get (Holder : Handle) return Object is
   begin
      return Holder.Ptr;
   end Get;

   procedure List_Append (List : Handle; Item : Handle) is
   begin
      Check_Handle (List);
      if Links.List_Append (List.Ptr, Item.Ptr) = -1 then
         Check_Error;
      end if;
   end List_Append;

   function List_New (Size : ssize_t) return Handle is
   begin
      return (Ada.Finalization.Controlled with Links.List_New (Size));
   end List_New;

   function List_GetItem
            (  List  : Handle;
               Index : ssize_t
            )  return Handle is
      Result : Handle;
   begin
      Check_Handle (List);
      Result.Ptr := Links.List_GetItem (List.Ptr, Index);
      if Result.Ptr = Null_Object then
         Check_Error;
      else
         Links.IncRef (Result.Ptr); -- Borrowed reference
      end if;
      return Result;
   end List_GetItem;

   procedure List_Insert
             (  List  : Handle;
                Index : ssize_t;
                Item  : Handle
             )  is
   begin
      Check_Handle (List);
      if Links.List_Insert (List.Ptr, Index, Item.Ptr) = -1 then
         Check_Error;
      end if;
   end List_Insert;

   function List_Size (Object : Handle) return ssize_t is
   begin
      if Object.Ptr = Null_Object then
         return 0;
      else
         return Links.List_Size (Object.Ptr);
      end if;
   end List_Size;

   function Long_AsLong (Object : Handle) return long is
      Result : constant long := Links.Long_AsLong (Object.Ptr);
   begin
      Check_Error;
      return Result;
   end Long_AsLong;

   function Long_FromLong (Value : long) return Handle is
   begin
      return
      (  Ada.Finalization.Controlled
      with
         Links.Long_FromLong (Value)
      );
   end Long_FromLong;

   function Module_Create
            (  Modules : ModuleDef;
               Version : int
            )  return Object is
   begin
      return Links.Module_Create (Modules, Version);
   end Module_Create;

   function Module_GetDict (Module : Handle) return Handle is
      Result : Handle;
   begin
      if Module.Ptr /= Null_Object then
         Result.Ptr := Links.Module_GetDict (Module.Ptr);
         if Result.Ptr = Null_Object then
            Check_Error;
         else
            Links.IncRef (Result.Ptr); -- Borrowed reference
         end if;
      end if;
      return Result;
   end Module_GetDict;

   function No_Value return Handle is
      Result : Handle;
   begin
      Result.Ptr := Links.None;
      Links.IncRef (Result.Ptr);
      return Result;
   end No_Value;

   function Object_CallObject
            (  Operation : Handle;
               Arguments : Handle
            )  return Handle is
   begin
      Check_Handle (Operation);
      return
      (  Ada.Finalization.Controlled
      with
         Links.Object_CallObject (Operation.Ptr, Arguments.Ptr)
      );
   end Object_CallObject;

   function Object_GetAttrString
            (  Module : Handle;
               Name   : String
            )  return Handle is
      Result : Handle;
   begin
      Check_Handle (Module);
      Result.Ptr := -- New reference
         Links.Object_GetAttrString (Module.Ptr, To_C (Name));
      if Result.Ptr = Null_Object then
         Check_Error;
      end if;
      return Result;
   end Object_GetAttrString;

   function Object_Dir (Value : Handle) return Handle is
       Result : Handle;
   begin
      if Value.Ptr /= Null_Object then
         Result.Ptr := Links.Object_Dir (Value.Ptr);
         if Result.Ptr = Null_Object then
            Check_Error;
         end if;
      end if;
      return Result;
   end Object_Dir;

   function Object_Str (Value : Object) return String is
   begin
      if Value = Null_Object then
         return "";
      end if;
      declare
         Result : constant Object := Links.Object_Str (Value);
      begin
         if Result = Null_Object then
            return "";
         else
            return As_String
                   (  (Ada.Finalization.Controlled with Result)
                   );
         end if;
      end;
   end Object_Str;

   function Object_Str (Value : Handle) return String is
   begin
      if Value.Ptr = Null_Object then
         return "";
      else
         return Object_Str (Value.Ptr);
      end if;
   end Object_Str;

   function Quit (Data : System.Address) return int is
   begin
      Links.Err_SetString (Links.Exc_SystemError, "Abort" & Nul);
      return 0;
   end Quit;

   procedure Request_Abort is
   begin
      if 0 = Links.AddPendingCall (Quit'Access, System.Null_Address)
      then
         null;
      end if;
   end Request_Abort;

   procedure Reraise_As (ID : Exception_ID) is
      Error : Object;
   begin
      Error := Links.Err_Occurred.all;
      if Error /= Null_Object then
         declare
            Class   : Handle;
            Message : Handle;
            Trace   : Handle;
         begin
            Links.Err_Fetch (Class.Ptr, Message.Ptr, Trace.Ptr);
            Links.Err_NormalizeException
            (  Class.Ptr,
               Message.Ptr,
               Trace.Ptr
            );
            Raise_Exception (ID, Object_Str (Message));
         end;
      end if;
   end Reraise_As;

   procedure Sequence_DelItem (List : Handle; Index : ssize_t) is
   begin
      Check_Handle (List);
      if 0 = Links.Sequence_DelItem (List.Ptr, Index) then
         null;
      end if;
   end Sequence_DelItem;

   function Sequence_Index
            (  List : Handle;
               Item : Handle
            )  return ssize_t is
      Result : ssize_t;
   begin
      Check_Handle (List);
      Check_Handle (Item);
      Result := Links.Sequence_Index (List.Ptr, Item.Ptr);
      if Result < 0 then
         Check_Error;
      end if;
      return Result;
   end Sequence_Index;

   function Sys_GetObject (Name : String) return Handle is
      Result : Handle;
   begin
      Result.Ptr := Links.Sys_GetObject (To_C (Name));
      if Result.Ptr /= Null_Object then
         Links.IncRef (Result.Ptr); -- Borrowed reference
      end if;
      return Result;
   end Sys_GetObject;

   procedure Throw_BadArgument is
   begin
      if 0 = Links.Err_BadArgument.all then
         null;
      end if;
   end Throw_BadArgument;

--     procedure Throw_KeyError (Message : String) is
--     begin
--        Links.Err_SetString (Links.Exc_KeyError, To_C (Message));
--     end Throw_KeyError;
--
--     procedure Throw_RuntimeError
--               (  Message : String;
--                  Error   : Exception_Occurrence
--               )  is
--     begin
--        Links.Err_SetString
--        (  Links.Exc_RuntimeError,
--           To_C (Message & Exception_Information (Error))
--        );
--     end Throw_RuntimeError;

   procedure Throw_TypeError (Message : String) is
   begin
      if TypeError = Null_Object then
         if 0 = Links.Err_BadArgument.all then
            null;
         end if;
         declare
            Value : Object;
            Trace : Object;
         begin
            Links.Err_Fetch (TypeError, Value, Trace);
            if Value /= Null_Object then
               Links.DecRef (Value);
            end if;
            Value := Links.Unicode_FromString (To_C (Message));
            Links.Err_Restore (TypeError, Value, Trace);
         end;
      else
         Links.Err_SetString (TypeError, To_C (Message));
      end if;
   end Throw_TypeError;

   procedure Throw_SystemError
             (  Message : String;
                Error   : Exception_Occurrence
             )  is
   begin
      if SystemError = Null_Object then
         Links.Err_BadInternalCall.all;
         declare
            Value : Object;
            Trace : Object;
         begin
            Links.Err_Fetch (SystemError, Value, Trace);
            if Value /= Null_Object then
               Links.DecRef (Value);
            end if;
            Value := Links.Unicode_FromString (To_C (Message));
            Links.Err_Restore (SystemError, Value, Trace);
         end;
      else
         Links.Err_SetString (SystemError, To_C (Message));
      end if;
   end Throw_SystemError;

--     procedure Throw_ValueError (Message : String) is
--     begin
--        Links.Err_SetString (Links.Exc_ValueError, To_C (Message));
--     end Throw_ValueError;

   function Tuple_GetItem
             (  Tuple    : Handle;
                Position : ssize_t
             )  return Handle is
      Result : Handle;
   begin
      Check_Handle (Tuple);
      Result.Ptr := Links.Tuple_GetItem (Tuple.Ptr, Position);
      if Result.Ptr = Null_Object then
         Check_Error;
      else
         Links.IncRef (Result.Ptr); -- Borrowed reference
      end if;
      return Result;
   end Tuple_GetItem;

   function Tuple_New (Size : ssize_t) return Handle is
      Result : Handle;
   begin
      Result.Ptr := Links.Tuple_New (Size);
      if Result.Ptr = Null_Object then
         Check_Error;
      end if;
      return Result;
   end Tuple_New;

   procedure Tuple_SetItem
              (  Tuple    : Handle;
                 Position : ssize_t;
                 Item     : Handle
              )  is
   begin
      Check_Handle (Tuple);
      if Links.Tuple_SetItem (Tuple.Ptr, Position, Item.Ptr) = 0  then
         Links.IncRef (Item.Ptr); -- Success, reference stolen
      else
         Check_Error;
      end if;
   end Tuple_SetItem;

   function Tuple_Size (Tuple : Handle) return ssize_t is
      Result : ssize_t := 0;
   begin
      if Tuple.Ptr /= Null_Object then
         Result := Links.Tuple_Size (Tuple.Ptr);
         Check_Error;
      end if;
      return Result;
   end Tuple_Size;

   function Unicode_DecodeFSDefault (Name : String) return Handle is
      Result : Handle;
   begin
      Result.Ptr := Links.Unicode_DecodeFSDefault (To_C (Name));
      if Result.Ptr = Null_Object then -- New reference
         Check_Error;
      end if;
      return Result;
   end Unicode_DecodeFSDefault;

   function Unicode_FromString (Name : String) return Handle is
      Result : Handle;
   begin
      Result.Ptr := Links.Unicode_FromString (To_C (Name));
      if Result.Ptr = Null_Object then -- New reference
         Check_Error;
      end if;
      return Result;
   end Unicode_FromString;

   function Unicode_GetLength (Object : Handle) return ssize_t is
      Result : ssize_t := 0;
   begin
      if Object.Ptr /= Null_Object then
         Result := Links.Unicode_GetLength (Object.Ptr);
         Check_Error;
      end if;
      return Result;
   end Unicode_GetLength;

   function Unicode_ReadChar (Object : Handle; Index : ssize_t)
      return Wide_Wide_Character is
      Result : Wide_Wide_Character;
   begin
      Check_Handle (Object);
      Result := Links.Unicode_ReadChar (Object.Ptr, Index);
      Check_Error;
      return Result;
   end Unicode_ReadChar;

end Py;
