--                                                                    --
--  package Py                      Copyright (c)  Dmitry A. Kazakov  --
--  Interface                                      Luebeck            --
--                                                 Winter, 2018       --
--                                                                    --
--                                Last revision :  13:34 16 Dec 2018  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--

with Ada.Exceptions;        use Ada.Exceptions;
with Interfaces.C;          use Interfaces.C;
with Interfaces.C.Strings;  use Interfaces.C.Strings;
with System;                use System;

with Ada.Finalization;
with Interfaces.C.Pointers;

package Py is

   Python_Error : exception;

   procedure Load (Name : String := "");
   type ssize_t is new ptrdiff_t;

   type Global_Interpreter_Lock is
      new Ada.Finalization.Limited_Controlled with private;
   procedure Finalize (Lock : in out Global_Interpreter_Lock);
   procedure Initialize (Lock : in out Global_Interpreter_Lock);

   type Handle is new Ada.Finalization.Controlled with private;
   procedure Adjust (Object : in out Handle);
   function As_String (Object : Handle) return String;
   function Bytes_AsString (Object : Handle) return String;
   function Callable_Check (Object : Handle) return int;
--
-- Check_Error -- Raise Python_Error if a Python exception is pending
--
   procedure Check_Error;
   function Dict_GetItemString
            (  Dictionary : Handle;
               Key        : String
            )  return Handle;
   function Dict_Keys (Dictionary : Handle) return Handle;
   function Dict_Items (Dictionary : Handle) return Handle;
   function Dict_New return Handle;
   procedure Dict_SetItemString
             (  Dictionary : Handle;
                Key        : String;
                Value      : Handle'Class
             );
   function Dict_Size (Dictionary : Handle) return ssize_t;
   procedure Err_Clear;
   function Err_Occurred return Handle;
   procedure Error_Traceback
             (  Destination : in out String;
                Pointer     : in out Integer;
                No_Error    : out Boolean;
                Decorator   : String := Character'Val (13) &
                                        Character'Val (10) &
                                        "   "
             );
   procedure Eval_InitThreads;
   procedure Finalize (Object : in out Handle);
   function FinalizeEx return int;
   procedure Initialize;
   procedure Invalidate (Object : in out Handle);
   function Import_Import (Name : Handle) return Handle;
   procedure Import_Import
             (  Name        : Handle;
                Module      : out Handle;
                Destination : in out String;
                Pointer     : in out Integer;
                No_Error    : out Boolean;
                Decorator   : String := Character'Val (13) &
                                        Character'Val (10) &
                                        "   "
             );
   procedure Import_Import
             (  Name        : String;
                Module      : out Handle;
                Destination : in out String;
                Pointer     : in out Integer;
                No_Error    : out Boolean;
                Decorator   : String := Character'Val (13) &
                                        Character'Val (10) &
                                        "   "
             );
   function Import_ImportModule (Name : String) return Handle;
   function Is_None (Object : Handle) return Boolean;
   function Is_Valid (Object : Handle) return Boolean;
   procedure List_Append (List : Handle; Item : Handle);
   function List_GetItem (List : Handle; Index : ssize_t) return Handle;
   procedure List_Insert
             (  List  : Handle;
                Index : ssize_t;
                Item  : Handle
             );
   function List_New (Size : ssize_t) return Handle;
   function List_Size (Object : Handle) return ssize_t;
   function Long_AsLong (Object : Handle) return long;
   function Long_FromLong (Value : long) return Handle;
   function Module_GetDict (Module : Handle) return Handle;
   function No_Value return Handle;
   function Object_CallObject
            (  Operation : Handle;
               Arguments : Handle
            )  return Handle;
   function Object_GetAttrString
            (  Module : Handle;
               Name   : String
            )  return Handle;
   function Object_Dir (Value : Handle) return Handle;
   function Object_Str (Value : Handle) return String;
   procedure Request_Abort;
--
-- Reraise_As -- Convert a Python exception into Ada exception
--
--    ID - The Ada exception to raise
--
-- If a Python  exception is pending it is caught  and  ID exception  is
-- raised instead
--
   procedure Reraise_As (ID : Exception_ID);
   procedure Sequence_DelItem (List : Handle; Index : ssize_t);
   function Sequence_Index
            (  List : Handle;
               Item : Handle
            )  return ssize_t;
   function Sys_GetObject (Name : String) return Handle;
--
-- Throw_BadArgument -- Set Python exception TypeError
--
   procedure Throw_BadArgument;
--
-- Throw_TypeError -- Set Python exception TypeError
--
--    Message - The message text
--
   procedure Throw_TypeError (Message : String);
--
-- Throw_SystemError -- Convert Ada exception RuntimeError
--
--    Message - The message text
--    Error   - Ada exception
--
    procedure Throw_SystemError
              (  Message : String;
                 Error   : Exception_Occurrence
              );
   function Tuple_GetItem
            (  Tuple    : Handle;
               Position : ssize_t
            )  return Handle;
   function Tuple_New (Size : ssize_t) return Handle;
   procedure Tuple_SetItem
             (  Tuple    : Handle;
                Position : ssize_t;
                Item     : Handle
             );
   function Tuple_Size (Tuple : Handle) return ssize_t;
   function Unicode_DecodeFSDefault (Name : String) return Handle;
   function Unicode_FromString (Name : String) return Handle;
   function Unicode_GetLength (Object : Handle) return ssize_t;
   function Unicode_ReadChar (Object : Handle; Index : ssize_t)
      return Wide_Wide_Character;

private
   type GILState is new System.Address;

   type Object is new Address;
   Null_Object : constant Object := Object (Null_Address);

   type Handle is new Ada.Finalization.Controlled with record
      Ptr : Object := Null_Object;
   end record;

   METH_VARARGS  : constant := 16#0001#;
   METH_KEYWORDS : constant := 16#0002#;
   METH_NOARGS   : constant := 16#0004#;
   METH_O        : constant := 16#0008#;

   type Object_Head is record
      Object_Count : ssize_t;
      Object_Type  : Object;
   end record;
   pragma Convention (C, Object_Head);

   type ModuleDef_Base is record
      Base  : Object_Head;
      Init  : access function return Object;
      Index : ssize_t;
      Copy  : Object;
   end record;
   pragma Convention (C, ModuleDef_Base);

   type CFunction is access function
        (  Self : Object;
           Args : Object
        )  return Object;
   pragma Convention (C, CFunction);

   type CFunctionWithKeywords is access function
        (  Self     : Object;
           Args     : Object;
           Keywords : Object
        )  return Object;
   pragma Convention (C, CFunctionWithKeywords);

   type Method_Entry (Keywords : Boolean := False) is record
      case Keywords is
         when False =>
            Positional : CFunction;
         when True =>
            Keyed : CFunctionWithKeywords;
      end case;
   end record;
   pragma Unchecked_Union (Method_Entry);
   pragma Convention (C, Method_Entry);

   type MethodDef is record
      Name  : chars_ptr;    -- The name of the built-in function/method
      Meth  : Method_Entry; -- The C function that implements it
      Flags : int;        -- Combination of METH_xxx flags, which mostly
                          -- describe the args expected by the C func
      Doc   : chars_ptr;  -- The __doc__ attribute, or NULL
   end record;
   pragma Convention (C, MethodDef);

   type ModuleDef_Slot is record
      Slot  : int;
      Value : System.Address;
   end record;
   pragma Convention (C, ModuleDef_Slot);

   type Visitproc is access function
        (  This : Object;
           Data : System.Address
        )  return int;
   pragma Convention (C, Visitproc);

   type Traverseproc is access function
        (  This  : Object;
           Visit : Visitproc;
           Data  : System.Address
        )  return int;
   pragma Convention (C, Traverseproc);

   type Inquiry is access function (This : Object) return int;
   pragma Convention (C, Inquiry);

   type Freefunc is access procedure (Address : System.Address);
   pragma Convention (C, Freefunc);

   type ModuleDef is record
      Base     : ModuleDef_Base;
      Name     : chars_ptr;
      Doc      : chars_ptr;
      Size     : ssize_t;
      Methods  : access MethodDef;
      Slots    : access ModuleDef_Slot;
      Traverse : Traverseproc;
      Clear    : Inquiry;
      Free     : Freefunc;
   end record;
   pragma Convention (C, ModuleDef);

   function Module_Create
            (  Modules : ModuleDef;
               Version : int
            )  return Object;

   type InitTab is access function return Object;
   pragma Convention (C, InitTab);

   function Import_AppendInittab
            (  Name : char_array;
               Init : InitTab
            )  return int;

   function Object_Str (Value : Object) return String;

   package Char_Ptrs is
       new Interfaces.C.Pointers
           (  Index              => size_t,
              Element            => char,
              Element_Array      => char_array,
              Default_Terminator => nul
           );

   function Is_Null (Reference : Object) return Boolean;

   type Pending_Call_Ptr is access function
        (  Data : System.Address
        )  return int;
   pragma Convention (C, Pending_Call_Ptr);

   function Quit (Data : System.Address) return int;
   pragma Convention (C, Quit);
------------------------------------------------------------------------
--
-- Python DLL entry points
--
   type AddPendingCall_Ptr is access function
        (  Call : Pending_Call_Ptr;
           Data : System.Address
        )  return int;
   pragma Convention (C, AddPendingCall_Ptr);

   type Bytes_AsStringAndSize_Ptr is access function
        (  Reference : Object;
           Buffer    : access Char_Ptrs.Pointer;
           Size      : access ssize_t
        )  return int;
   pragma Convention (C, Bytes_AsStringAndSize_Ptr);

   type Callable_Check_Ptr is access
      function (Reference : Object) return int;
   pragma Convention (C, Callable_Check_Ptr);

   type Dict_GetItemString_Ptr is access function
        (  Dictionary : Object;
           Key        : char_array
        )  return Object;
   pragma Convention (C, Dict_GetItemString_Ptr);

   type Dict_Items_Ptr is access function
        (  Dictionary : Object
        )  return Object;
   pragma Convention (C, Dict_Items_Ptr);

   type Dict_Keys_Ptr is access function
        (  Dictionary : Object
        )  return Object;
   pragma Convention (C, Dict_Keys_Ptr);

   type Dict_New_Ptr is access function return Object;
   pragma Convention (C, Dict_New_Ptr);

   type Dict_SetItemString_Ptr is access function
        (  Dictionary : Object;
           Key        : char_array;
           Value      : Object
        )  return int;
   pragma Convention (C, Dict_SetItemString_Ptr);

   type Dict_Size_Ptr is access function
        (  Dictionary : Object
        )  return ssize_t;
   pragma Convention (C, Dict_Size_Ptr);

   type DecRef_Ptr is access procedure (Reference : Object);
   pragma Convention (C, DecRef_Ptr);

   type Err_BadArgument_Ptr is access function return int;
   pragma Convention (C, Err_BadArgument_Ptr);

   type Err_BadInternalCall_Ptr is access procedure;
   pragma Convention (C, Err_BadInternalCall_Ptr);

   type Err_Clear_Ptr is access procedure;
   pragma Convention (C, Err_Clear_Ptr);

   type Err_Occurred_Ptr is access function return Object;
   pragma Convention (C, Err_Occurred_Ptr);

   type Err_Fetch_Ptr is access procedure
        (  Error_Type : out Object;
           Value      : out Object;
           Traceback  : out Object
        );
   pragma Convention (C, Err_Fetch_Ptr);

   type Err_NewException_Ptr is access function
        (  Name       : char_array;
           Base       : Object := Null_Object;
           Dictionary : Object := Null_Object
        )  return Object;
   pragma Convention (C, Err_NewException_Ptr);

   type Err_NormalizeException_Ptr is access procedure
        (  Error_Type : in out Object;
           Value      : in out Object;
           Traceback  : in out Object
        );
   pragma Convention (C, Err_NormalizeException_Ptr);

   type Err_Restore_Ptr is access procedure
        (  Error_Type : Object;
           Value      : Object;
           Traceback  : Object
        );
   pragma Convention (C, Err_Restore_Ptr);

   type Err_SetString_Ptr is access procedure
        (  Error   : Object;
           Message : char_array
        );
   pragma Convention (C, Err_SetString_Ptr);

   type Eval_InitThreads_Ptr is access procedure;
   pragma Convention (C, Eval_InitThreads_Ptr);

   type FinalizeEx_Ptr is access function return int;
   pragma Convention (C, FinalizeEx_Ptr);

   type Float_AsDouble_Ptr is access function
        (  Value : Object
        )  return double;
   pragma Convention (C, Float_AsDouble_Ptr);

   type Float_FromDouble_Ptr is access function
        (  Value : double
        )  return Object;
   pragma Convention (C, Float_FromDouble_Ptr);

   type Get_Object_Ptr is access function return Object;
   pragma Convention (C, Get_Object_Ptr);

   type GILState_Ensure_Ptr is access function return GILState;
   pragma Convention (C, GILState_Ensure_Ptr);

   type GILState_Release_Ptr is access procedure (State : GILState);
   pragma Convention (C, GILState_Release_Ptr);

   type Import_AppendInittab_Ptr is access function
        (  Name : char_array;
           Init : InitTab
        )  return int;
   pragma Convention (C, Import_AppendInittab_Ptr);

   type IncRef_Ptr is access procedure (Reference : Object);
   pragma Convention (C, IncRef_Ptr);

   type InitializeEx_Ptr is access procedure (Init_Sigs : int);
   pragma Convention (C, InitializeEx_Ptr);

   type Import_Import_Ptr is access
      function (Name : Object) return Object;
   pragma Convention (C, Import_Import_Ptr);

   type Import_ImportModule_Ptr is access function
        (  Name : char_array
        )  return Object;
   pragma Convention (C, Import_ImportModule_Ptr);

   type List_Append_Ptr is access function
        (  List : Object;
           Item : Object
        )  return int;
   pragma Convention (C, List_Append_Ptr);

   type List_GetItem_Ptr is access function
        (  List  : Object;
           Index : ssize_t
        )  return Object;
   pragma Convention (C, List_GetItem_Ptr);

   type List_Insert_Ptr is access function
        (  List  : Object;
           Index : ssize_t;
           Item  : Object
        )  return int;
   pragma Convention (C, List_Insert_Ptr);

   type List_New_Ptr is access function (Size : ssize_t) return Object;
   pragma Convention (C, List_New_Ptr);

   type List_Size_Ptr is access function
        (  List : Object
        )  return ssize_t;
   pragma Convention (C, List_Size_Ptr);

   type Long_AsLong_Ptr is access function
        (  Reference : Object
        )  return long;
   pragma Convention (C, Long_AsLong_Ptr);

   type Long_FromLong_Ptr is access function
        (  Value : long
        )  return Object;
   pragma Convention (C, Long_FromLong_Ptr);

   type Module_Create_Ptr is access function
        (  Definition : ModuleDef;
           Version    : int
        )  return Object;
   pragma Convention (C, Module_Create_Ptr);

   type Module_GetDict_Ptr is access function
        (  Mpdule : Object
        )  return Object;
   pragma Convention (C, Module_GetDict_Ptr);

   type Object_CallObject_Ptr is access
      function (Operation : Object; Arguments : Object) return Object;
   pragma Convention (C, Object_CallObject_Ptr);

   type Object_Dir_Ptr is access
      function (Thing : Object) return Object;
   pragma Convention (C, Object_Dir_Ptr);

   type Object_GetAttrString_Ptr is access function
        (  Module : Object;
           Name   : char_array
        )  return Object;
   pragma Convention (C, Object_GetAttrString_Ptr);

   type Object_Str_Ptr is access function
        (  Value : Object
        )  return Object;
   pragma Convention (C, Object_Str_Ptr);

   type Sequence_DelItem_Ptr is access function
        (  List  : Object;
           Index : ssize_t
        )  return int;
   pragma Convention (C, Sequence_DelItem_Ptr);

   type Sequence_Index_Ptr is access function
        (  List : Object;
           Item : Object
        )  return ssize_t;
   pragma Convention (C, Sequence_Index_Ptr);

   type Sys_GetObject_Ptr is access function
        (  Name : char_array
        )  return OBject;
   pragma Convention (C, Sys_GetObject_Ptr);

   type Tuple_GetItem_Ptr is access function
        (  Tuple    : Object;
           Position : ssize_t
        )  return Object;
   pragma Convention (C, Tuple_GetItem_Ptr);

   type Tuple_New_Ptr is access function (Size : ssize_t) return Object;
   pragma Convention (C, Tuple_New_Ptr);

   type Tuple_SetItem_Ptr is access function
        (  Tuple    : Object;
           Position : ssize_t;
           Item     : Object
        )  return int;
   pragma Convention (C, Tuple_SetItem_Ptr);

   type Tuple_Size_Ptr is access
      function (Tuple : Object) return ssize_t;
   pragma Convention (C, Tuple_Size_Ptr);

   type Unicode_AsEncodedString_Ptr is access function
        (  Unicode  : Object;
           Encoding : char_array := "utf-8" & Nul;
           Errors   : char_array := "Error ~" & Nul
        )  return Object;
   pragma Convention (C, Unicode_AsEncodedString_Ptr);

   type Unicode_DecodeFSDefault_Ptr is access
      function (Name : char_array) return Object;
   pragma Convention (C, Unicode_DecodeFSDefault_Ptr);

   type Unicode_FromString_Ptr is access
      function (Name : char_array) return Object;
   pragma Convention (C, Unicode_FromString_Ptr);

   type Unicode_GetLength_Ptr is access
      function (Value : Object) return ssize_t;
   pragma Convention (C, Unicode_GetLength_Ptr);

   type Unicode_ReadChar_Ptr is access function
        (  Value : Object;
           Index : ssize_t
        )  return Wide_Wide_Character;
   pragma Convention (C, Unicode_ReadChar_Ptr);

   package Links is
      AddPendingCall            : AddPendingCall_Ptr;
      Bytes_AsStringAndSize     : Bytes_AsStringAndSize_Ptr;
      Callable_Check            : Callable_Check_Ptr;
      DecRef                    : DecRef_Ptr;
      Dict_New                  : Dict_New_Ptr;
      Dict_GetItemString        : Dict_GetItemString_Ptr;
      Dict_Items                : Dict_Items_Ptr;
      Dict_Keys                 : Dict_Keys_Ptr;
      Dict_SetItemString        : Dict_SetItemString_Ptr;
      Dict_Size                 : Dict_Size_Ptr;
      Err_BadArgument           : Err_BadArgument_Ptr;
      Err_BadInternalCall       : Err_BadInternalCall_Ptr;
      Err_Clear                 : Err_Clear_Ptr;
      Err_Fetch                 : Err_Fetch_Ptr;
      Err_NewException          : Err_NewException_Ptr;
      Err_NormalizeException    : Err_NormalizeException_Ptr;
      Err_Occurred              : Err_Occurred_Ptr;
      Err_Restore               : Err_Restore_Ptr;
      Err_SetString             : Err_SetString_Ptr;
      Eval_InitThreads          : Eval_InitThreads_Ptr;
      Exc_KeyError              : Object := Null_Object;
      Exc_RuntimeError          : Object := Null_Object;
      Exc_ValueError            : Object := Null_Object;
      Exc_SystemError           : Object := Null_Object;
      Exc_TypeError             : Object := Null_Object;
      False                     : Object := Null_Object;
      FinalizeEx                : FinalizeEx_Ptr;
      Float_AsDouble            : Float_AsDouble_Ptr;
      Float_FromDouble          : Float_FromDouble_Ptr;
      GILState_Ensure           : GILState_Ensure_Ptr;
      GILState_Release          : GILState_Release_Ptr;
      Import_AppendInittab      : Import_AppendInittab_Ptr;
      IncRef                    : IncRef_Ptr;
      InitializeEx              : InitializeEx_Ptr;
      Import_Import             : Import_Import_Ptr;
      Import_ImportModule       : Import_ImportModule_Ptr;
      List_Append               : List_Append_Ptr;
      List_GetItem              : List_GetItem_Ptr;
      List_Insert               : List_Insert_Ptr;
      List_New                  : List_New_Ptr;
      List_Size                 : List_Size_Ptr;
      Long_AsLong               : Long_AsLong_Ptr;
      Long_FromLong             : Long_FromLong_Ptr;
      Module_Create             : Module_Create_Ptr;
      Module_GetDict            : Module_GetDict_Ptr;
      None                      : Object := Null_Object;
      Object_CallObject         : Object_CallObject_Ptr;
      Object_Dir                : Object_Dir_Ptr;
      Object_GetAttrString      : Object_GetAttrString_Ptr;
      Object_Str                : Object_Str_Ptr;
      Sequence_DelItem          : Sequence_DelItem_Ptr;
      Sequence_Index            : Sequence_Index_Ptr;
      Sys_GetObject             : Sys_GetObject_Ptr;
      True                      : Object := Null_Object;
      Tuple_GetItem             : Tuple_GetItem_Ptr;
      Tuple_New                 : Tuple_New_Ptr;
      Tuple_SetItem             : Tuple_SetItem_Ptr;
      Tuple_Size                : Tuple_Size_Ptr;
      Unicode_AsEncodedString   : Unicode_AsEncodedString_Ptr;
      Unicode_DecodeFSDefault   : Unicode_DecodeFSDefault_Ptr;
      Unicode_FromString        : Unicode_FromString_Ptr;
      Unicode_GetLength         : Unicode_GetLength_Ptr;
      Unicode_ReadChar          : Unicode_ReadChar_Ptr;
   end Links;
   procedure Check_Links (Library_Name : String);

   function As_String (Value : Object) return String;

   type Global_Interpreter_Lock is
      new Ada.Finalization.Limited_Controlled with
   record
      State : GILState;
   end record;

--  --
--  -- Throw_KeyError -- Set Python exception KeyError
--  --
--  --    Message - The message text
--  --
--     procedure Throw_KeyError (Message : String);
--  --
--  -- Throw_RuntimeError -- Convert Ada exception RuntimeError
--  --
--  --    Message - The message text
--  --    Error   - Ada exception
--  --
--     procedure Throw_RuntimeError
--               (  Message : String;
--                  Error   : Exception_Occurrence
--               );
--  --
--  -- Throw_ValueError -- Set Python exception ValueError
--  --
--  --    Message - The message text
--  --
--     procedure Throw_ValueError (Message : String);

end Py;
